#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2023/3/1  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function, unicode_literals)

class RegisterProviders():
    def bootstrap(self, core):
        providers = core.make("config").get("providers")
        if not providers:
            return

        for provider in providers:
            (core.make(provider, core)).register()


# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2023/2/2
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function, unicode_literals)


from noba.snippet import *

__all__ = ['Container', 'container']


class Container(object):
    """
    ioc ioccontainer
    """
    def __init__(self):
        self._binds = dict()
        self._instances = dict()
        self._with = list()
        self.aliases = dict()

    def __getitem__(self, item):
        return self.make(item)

    def __call__(self, abstract=None, parameters=[]):
        return self.make(abstract, parameters)

    def instance(self, abstract, instance):
        self._instances[abstract] = instance
        return instance

    def singleton(self, abstract, concrete=None):
        self.bind(abstract, concrete, True)

    def bind(self, abstract, concrete=None, shared=False):
        if not is_function(concrete):
            concrete = self.get_function(abstract, concrete)

        self._binds[abstract] = {
            'concrete': concrete,
            'shared': shared,
        }

    def bound(self, abstract):
        return (abstract in self._binds) or (abstract in self._instances)

    def get_function(self, abstract, concrete):
        def doit(container, parameters=[]):
            if abstract == concrete:
                return container.build(concrete);
            else:
                return container.make(concrete, parameters)
        return doit

    def make_with(self, abstract, parameters = []):
        return self.make(abstract, parameters);

    def make(self, abstract, parameters = []):
        is_alias = not is_list(abstract) and abstract in self.aliases
        abstract = self.get_alias(abstract)
        is_share_in_alias = False

        if is_alias:
            is_share_in_alias = True
        if is_list(abstract):
            abstract, is_share_in_alias = abstract

        needsContextualBuild = not (not parameters);
        isclass = False
        if is_class(abstract):
            abstract_cls = abstract
            package_name = abstract.__module__
            abstract = package_name + "." + abstract.__name__
            isclass = True

        # if singleton, return obj directly
        if (abstract in self._instances) and (not needsContextualBuild):
            return self._instances[abstract]

        parameters and (self._with).append(parameters)

        concrete = self.get_concrete(abstract, abstract_cls if isclass else None)

        if self.is_buildable(concrete, abstract):
            obj = self.build(concrete)
        else:
            obj = self.make(concrete)

        if self.is_shared(abstract, is_share_in_alias):
            self._instances[abstract] = obj

        return obj

    def is_buildable(self, concrete, abstract):
        return  (concrete == abstract) or is_function(concrete)

    def get_concrete(self, abstract, abstract_cls):
        if abstract in self._binds:
            return self._binds[abstract]['concrete']

        if abstract_cls:
            for b in self._binds:
                if "." in b:
                    m = dyn_from_path_import(b)
                    if m == abstract_cls:
                        return self._binds[b]['concrete']

        return abstract

    def build(self, concrete):
        if is_function(concrete):
            if concrete.__name__ == "__init__":
                Last_parameter = self.get_Last_parameter_override()
                return concrete(self, Last_parameter)
            else:
                return concrete(self)

        if is_class(concrete):
            reflector = concrete
        elif is_str(concrete):
            reflector = self.get_class_by_name(concrete)
        else:
            raise TypeError(f"build need str or class, but {type(concrete)} were given")

        init_fun = reflector.__init__

        if not is_function(init_fun):
            return reflector()

        init_param = inspect.signature(reflector.__init__).parameters.values()
        dependencies: list = self.get_dependencies(init_param)
        if reflector.__name__ == 'ABCMeta':
            raise TypeError("")

        # if dependencies is empty，ignore it
        if is_all_empty(dependencies):
            return reflector()
        else:
            return reflector(*dependencies)

    def get_Last_parameter_override(self):
        last_with = (self._with.pop()) if (self._with) else  []
        return last_with


    def get_class_by_name(self, cls):
        try:
            return dyn_from_path_import(cls)
        except KeyError:
            return cls

    def get_alias(self, abstract):
        return self.get_alias(self.aliases[abstract]) if (not is_list(abstract) and abstract in self.aliases) else abstract

    def is_shared(self, abstract, is_share_in_alias):
        return (abstract in self._instances) or (is_share_in_alias) or (abstract in self._binds and self._binds[abstract]['shared'])

    def get_dependencies(self, parameters):
        dependencies = []
        for param in  list(parameters)[1:]:
            if not param.annotation is inspect._empty:
                if None:
                    dependencies.append(None)
                else:
                    dependencies.append(self.resolve_class(param.annotation))
            else:
                dependencies.append(self.get_Last_parameter_override())
        return dependencies

    def resolve_class(self, parameter):
        return self.make(parameter)

    def get_bindings(self):
        return self._binds

container = Container()


#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2023/3/5  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function, unicode_literals)
from noba.snippet import *

class EventHub:
    hubs = {}
    def __init__(self, event, name=None):
        if not is_list(event):
            event = [event]
        self._event = event
        self._listeners = list()
        if name and self._single(name):
            EventHub.hubs[name] = self

    def _single(self, name):
        if name in EventHub.hubs:
            print(f"[Error] EventHub '{name}' already exists!")
            exit()
        else:
            return True

    def name(self, name=None):
        if name and self._single(name):
            self._name = name
            EventHub.hubs[name] = self
        return self

    @classmethod
    def hub(cls, event, name=None):
        return cls(event, name)

    @classmethod
    def get_all_hub(cls):
        return cls.hubs

    @classmethod
    def drop(cls, name):
        del EventHub.hubs[name]

    @classmethod
    def get_hub(cls, name):
        try:
            return EventHub.hubs[name]
        except KeyError as e:
            print(f"EventHub '{name}' non-existent")
            exit()

    def get_events(self):
        return self._event

    def get_name(self):
        return get_key_by_value(EventHub.hubs, self)[0]

    @property
    def manager(self):
        return self.__class__

    def drop_me(self):
        del EventHub.hubs[self.get_name()]

    def watch(self, event, handler, always=False):
        if not event in self._event:
            print(f'[EventHub] watch，attempt to watch an invalid event：{event}. List of valid events: {self._event}')
            return

        self._listeners.append({
            'event': event,
            'always': always,
            'handler': handler,
            'trigger_count': 0,
            'limit_count': 1 if not always else  0
        })

    def fire(self, event, data=None):
        if not event in self._event:
            print(f'[EventHub] fire，attempt to trigger an invalid event：{event}. List of valid events: {self._event}')
            return

        for listener in self._listeners:
            if not listener['event'] == event:
                continue

            try:
                listener['handler'](data)

            except Exception as e:
                print(f"[EventHub] caught err when exec handler, err:, {e}, event:, {event}, handler:, {listener.handler}")

            listener['trigger_count'] += 1

        self._listeners = [listener for listener in self._listeners if not (listener['limit_count'] > 0 and listener['limit_count'] <= listener['trigger_count'])]

    def unwatch(self, event, handler):
        self._listeners = [listener for listener in self._listeners if not (listener['event'] == event)]












#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2023/2/28  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function, unicode_literals)
from functools import reduce
from noba.snippet import *
from noba.Core import core

class Pipeline:
    def __init__(self, container=core):
        self.container = container
        self.method = 'handle'
        self.pipes = list()
        self._pipeline_folder = 'pipeline'

    def folder(self, folder):
        self._pipeline_folder = folder
        return self

    def send(self, passable):
        self.passable = passable
        return self

    def through(self, pipes, *args):
        self.pipes = pipes if isinstance(pipes, list) else list(args)
        return self


    def pipe(self, pipes, *args):
        (self.pipes).extend(pipes if is_list(pipes) else list(args))
        return self

    def via(self, method):
        self.method = method
        return self

    def then(self, destination):
        pipeline = reduce(
            self.__carry(), self._pipes()[::-1], self._prepare_destination(destination)
        )

        return pipeline(self.passable)

    def then_return(self):
        return self.then(lambda passable:passable)


    def _prepare_destination(self, destination):
        def func(passable):
            try:
                return destination(passable)
            except Exception as e:
                return self._handle_exception(passable, e)
        return func

    def __carry(self):
        def one_func(stack, piper):
            def two_func(passable):
                pipe = piper
                try:
                    if (is_function(pipe)):
                        # If the pipe is a callable, then we will call it directly, but otherwise we
                        # will resolve the pipes out of the dependency container and call it with
                        # the appropriate method and arguments, returning the results back out.
                        return pipe(passable, stack)
                    elif (is_str(pipe)):
                        name, parameters_r = self._parse_pipe_string(pipe)
                        # If the pipe is a string we will parse the string and resolve the class out
                        # of the dependency injection container. We can then build a callable and
                        # execute the pipe function giving in the parameters that are required.
                        pipe = self._get_container().make(name)
                        # parameters = [passable, stack].extend(parameters)
                        parameters = [passable, stack]
                        parameters.extend(parameters_r)
                    else:
                        # If the pipe is already an object we'll just make a callable and pass it to
                        # the pipe as-is. There is no need to do any extra parsing and formatting
                        # since the object we're given was already a fully instantiated object.
                        parameters = [passable, stack]

                    # print(parameters)
                    carry = getattr(pipe, self.method)(*parameters) if hasattr(pipe, self.method) else pipe(*parameters)
                    return self._handle_carry(carry)

                except Exception as e:
                    return self._handle_exception(passable, e)

            return two_func
        return one_func

    def _parse_pipe_string(self, pipe):
        split = [x for x in pipe.split(':', 2)]

        if len(split) < 2:
            split.append([])
        name, parameters = split

        if is_str(parameters):
            parameters = parameters.split(',')

        if not self._pipeline_folder in name:
            name = self._pipeline_folder + '.' + name
        return [name, parameters]

    def _pipes(self):
        return self.pipes

    def _get_container(self):
        if (not self.container):
            raise RuntimeError('A container instance has not been passed to the Pipeline.')

        return self.container

    def set_container(self, container):
        self.container = container
        return self

    def _handle_carry(self, carry):
        return carry

    def _handle_exception(self, passable, e):
        raise e


#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2023/3/10  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function, unicode_literals)

# class BBService(object):
#     pass
#
#
# class BTService(object):
#     pass


from noba.confighandler import custom_config
import backtrader as bt
bt.custom_config = custom_config
import noba.tabs as tabs
#tabs: AnalyzerTab, MetadataTab, ConfigTab, LogTabs, SourceTab
bt.tabs = tabs

bt.tabs.LogTabs = lambda cols:type("LogTab",(bt.tabs.LogTab,),{'cols':cols})
bt.getlogger = bt.tabs.log.getlogger
bt.get_order_logger = lambda name="Order Trace": bt.getlogger(col=['Day', 'Ref', 'OrdType', 'Status', 'Size', 'Remsize', 'Alive'], name=name)


from noba.app import BacktraderBokeh
from noba.analyzers import LivePlotAnalyzer as BacktraderBokehLive
from noba.optbrowser import OptBrowser as BacktraderBokehOptBrowser
from noba.analyzers import RecorderAnalyzer
from noba.analyzers import TradelistAnalyzer
from noba.feeds import FakeFeed

import noba.schemes as schemes

# initialize analyzer tables
from noba.analyzer_tables import inject_datatables
inject_datatables()

from noba.plugin import *
bt.Bokeh = BacktraderBokeh

# Customized analyzers and feeds
bt.analyzers.Live = BacktraderBokehLive
bt.analyzers.Recorder = bt.analyzers.RecorderAnalyzer = RecorderAnalyzer
bt.analyzers.Tradelist = bt.analyzers.TradelistAnalyzer = TradelistAnalyzer
bt.feeds.FakeFeed = FakeFeed

bt.Opt = BacktraderBokehOptBrowser

#schemes: Blackly, Tradimo
bt.schemes = schemes
bt.schemes.Black = schemes.Blackly
bt.schemes.White = schemes.Tradimo


#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2023/3/1  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function, unicode_literals)

from functools import reduce

class RegisterAlias():
    def bootstrap(self, core):
        user_alias = core.make("config").get("aliases")
        if not user_alias:
            return

        package_aliase = core.aliases
        core.aliases = reduce(self.__merge, [package_aliase, user_alias])

    def __merge(self, package_aliase, user_alias):
        if not package_aliase:
            return user_alias

        for k in user_alias:
            if k in package_aliase and isinstance(package_aliase[k], dict) and isinstance(user_alias[k], dict):
                package_aliase[k] = self.__merge(package_aliase[k], user_alias[k])
            elif(user_alias[k] or (not user_alias[k] and user_alias[k] is None)):
                package_aliase[k] = user_alias[k]
        return package_aliase


#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2023/2/12  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function, unicode_literals)

import inspect, importlib, threading, os, json, pprint, re

DIRECTORY_SEPARATOR =  os.sep

def has_key(obj, k):
    o = obj
    arr = k.split('.')
    for one in arr:
        if (hasattr(o, '__iter__')  and (not one in o)) or not hasattr(o, '__iter__'):
            return False
        else:
            o = o[one]
    return True

def get_value(obj, k, default= None):
    o = obj
    arr = k.split('.')
    for one in arr:
        if (hasattr(o, '__iter__')  and (not one in o)) or not hasattr(o, '__iter__'):
            return default
        else:
            o = o[one]
    return o

def set_value(obj, k, value):
    o = obj
    arr = k.split('.')

    for one in arr:
        if (hasattr(o, '__iter__')  and (not one in o)) or not hasattr(o, '__iter__'):
            return None
        else:
            if one == arr[-1]:
                o[one] = value
            o = o[one]
    return {k:o}


def get_class_by_name(cls):
    try:
        return globals()[cls]
    except KeyError:
        return cls

def is_list(arg):
    return isinstance(arg, list)

def is_dict(arg):
    return isinstance(arg, dict)

def is_class(some):
    return inspect.isclass(some)


def is_str(some):
    return isinstance(some, str)

def is_number(some):
    if re.match(r'^[-+]?\d+(\.\d+)?$', some):
        return True
    return False


def is_function(some):
    return inspect.isfunction(some)

def is_all_empty(lst):
    if not isinstance(lst, list):
        return False
    elif len(lst) == 0:
        return True
    else:
        return all(is_all_empty(elem) for elem in lst)

def get_key_by_value(my_dict, value):
    return [k for k, v in my_dict.items() if v == value]

def singletonDecorator(cls, *args, **kwargs):
    instance = {}

    def wrapperSingleton(*args, **kwargs):
        if cls not in instance:
            instance[cls] = cls(*args, **kwargs)
        return instance[cls]

    return wrapperSingleton

class DefaultMethodDecorator(object):
    def __init__(self, fun):
        fun.__getattr__ = lambda self,key: self.default(key)
        self.__fun = fun

    def __call__(self):
        obj = self.__fun()
        return obj

def import_module(mod):
    return importlib.import_module(mod)

def dyn_from_import(mod: str, cls: str):
    import_modual = import_module(mod)
    return getattr(import_modual, cls)

def dyn_from_path_import(path: str):
    import_modual = import_module('.'.join(path.split(".")[:-1]))
    return getattr(import_modual, path.split(".")[-1])

def delay(cbname,delay,*argments):
    threading.Timer(delay,cbname,argments).start()


def load_json_file_if_not_exist_return_empty_dict(file, *args, **kwargs):
    try:
        with open(file, "r", encoding='utf-8') as f:
            json_text = f.read()
        jsontodict = json.loads(json_text, *args, **kwargs)
        return jsontodict
    except FileNotFoundError as e:
        return dict()

try:
    from shutil import copyfile
except ImportError:
    def copyfile(src, dst):
        with open(src,'r', encoding='utf-8') as fr:
            content = fr.read()
        with open(dst,'w', encoding='utf-8') as fw:
            fw.write(content)

def write_config_to_json_file(file, config_dict):
    config_json = json.dumps(config_dict, indent=4)
    with open(file, "w", encoding='utf-8') as f:
        f.write(config_json)

def here(something, do_print=True):
    '''Get the current file and line number in Python script.'''
    frameinfo = inspect.getframeinfo(inspect.currentframe().f_back)
    if do_print:
        print(f"{frameinfo.filename.split(DIRECTORY_SEPARATOR)[-1]}, {frameinfo.lineno}: ")
        pprint.pprint(something)
        print("\r\n")


#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2023/3/1  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function, unicode_literals)

import sys
from pathlib import Path
from noba.service import Core


aliases = {
    'config': 'noba.service.config.ConfigManager',
    'pipeline':'noba.service.pipeline.Pipeline',
    'db':'noba.service.abs.DBManagerAbs',
    'DB':'noba.service.abs.DBManagerAbs'
}


try:
    package_absolute_path = Path(__file__).absolute().parent
    project_absolute_path = Path.cwd()

    sys.path.append(str(project_absolute_path))
except:
    pass

core = Core(package_absolute_path, project_absolute_path, aliases)

config = core.make("config")

core.bootstrap(config.get("bootstrappers"))